"Implementation of the zero-iteration-loop operator."

import parso
from parso.python.tree import ForStmt

from .operator import Operator


class ZeroIterationForLoop(Operator):
    """An operator that modified for-loops to have zero iterations."""

    def mutation_positions(self, node):
        if isinstance(node, ForStmt):
            expr = node.children[3]
            yield (expr.start_pos, expr.end_pos)

    def mutate(self, node, index):
        """Modify the For loop to evaluate to None"""
        assert index == 0
        assert isinstance(node, ForStmt)

        empty_list = parso.parse(' []')
        node.children[3] = empty_list
        return node

    @classmethod
    def examples(cls):
        return (
            ('for i in rang(1,2): pass', 'for i in []: pass'),
        )



"""This is the body of the low-level worker tool.
"""

import difflib
import traceback

import cosmic_ray.mutating
import cosmic_ray.plugins
from cosmic_ray.testing import run_tests
from cosmic_ray.work_item import TestOutcome, WorkerOutcome, WorkResult


# pylint: disable=R0913
def worker(module_path,
           python_version,
           operator_name,
           occurrence,
           test_command,
           timeout):
    """Mutate the OCCURRENCE-th site for OPERATOR_NAME in MODULE_PATH, run the
    tests, and report the results.

    This is fundamentally the single-mutation-and-test-run process
    implementation.

    There are three high-level ways that a worker can finish. First, it could
    fail exceptionally, meaning that some uncaught exception made its way from
    some part of the operation to terminate the function. This function will
    intercept all exceptions and return it in a non-exceptional structure.

    Second, the mutation testing machinery may determine that there is no
    OCCURENCE-th instance for OPERATOR_NAME in the module under test. In this
    case there is no way to report a test result (i.e. killed, survived, or
    incompetent) so a special value is returned indicating that no mutation is
    possible.

    Finally, and hopefully normally, the worker will find that it can run a
    test. It will do so and report back the result - killed, survived, or
    incompetent - in a structured way.

    Args:
        module_name: The path to the module to mutate
        python_version: The version of Python to use when interpreting the code in `module_path`.
            A string of the form "MAJOR.MINOR", e.g. "3.6" for Python 3.6.x.
        operator_name: The name of the operator plugin to use
        occurrence: The occurrence of the operator to apply
        test_command: The command to execute to run the tests
        timeout: The maximum amount of time (seconds) to let the tests run

    Returns: A WorkResult

    Raises: This will generally not raise any exceptions. Rather, exceptions
        will be reported using the 'exception' result-type in the return value.

    """
    try:
        operator_class = cosmic_ray.plugins.get_operator(operator_name)
        operator = operator_class(python_version)

        with cosmic_ray.mutating.use_mutation(module_path, operator,
                                              occurrence) as (original_code,
                                                              mutated_code):
            if mutated_code is None:
                return WorkResult(worker_outcome=WorkerOutcome.NO_TEST)

            test_outcome, output = run_tests(test_command, timeout)

            diff = _make_diff(original_code, mutated_code, module_path)

            return WorkResult(
                output=output,
                diff='\n'.join(diff),
                test_outcome=test_outcome,
                worker_outcome=WorkerOutcome.NORMAL)

    except Exception:  # noqa # pylint: disable=broad-except
        return WorkResult(
            output=traceback.format_exc(),
            test_outcome=TestOutcome.INCOMPETENT,
            worker_outcome=WorkerOutcome.EXCEPTION)


def _make_diff(original_source, mutated_source, module_path):
    module_diff = ["--- mutation diff ---"]
    for line in difflib.unified_diff(
            original_source.split('\n'),
            mutated_source.split('\n'),
            fromfile="a" + str(module_path),
            tofile="b" + str(module_path),
            lineterm=""):
        module_diff.append(line)
    return module_diff



"A tool for creating XML reports."

import sys
import xml.etree.ElementTree

import docopt

from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.work_item import TestOutcome, WorkerOutcome


def report_xml():
    """cr-xml

Usage: cr-xml <session-file>

Print an XML formatted report of test results for continuous integration systems
"""
    arguments = docopt.docopt(report_xml.__doc__, version='cr-rate 1.0')
    with use_db(arguments['<session-file>'], WorkDB.Mode.open) as db:
        xml_elem = _create_xml_report(db)
        xml_elem.write(
            sys.stdout.buffer, encoding='utf-8', xml_declaration=True)


def _create_xml_report(db):
    errors = 0
    failed = 0
    skipped = 0
    root_elem = xml.etree.ElementTree.Element('testsuite')

    for work_item, result in db.completed_work_items:
        if result.worker_outcome in {
                WorkerOutcome.EXCEPTION, WorkerOutcome.ABNORMAL
        }:
            errors += 1
        if result.is_killed:
            failed += 1
        if result.worker_outcome == WorkerOutcome.SKIPPED:
            skipped += 1

        subelement = _create_element_from_work_item(work_item)
        subelement = _update_element_with_result(subelement, result)
        root_elem.append(subelement)

    for work_item in db.pending_work_items:
        subelement = _create_element_from_work_item(work_item)
        root_elem.append(subelement)

    root_elem.set('errors', str(errors))
    root_elem.set('failures', str(failed))
    root_elem.set('skips', str(skipped))
    root_elem.set('tests', str(db.num_work_items))
    return xml.etree.ElementTree.ElementTree(root_elem)


def _create_element_from_work_item(work_item):
    sub_elem = xml.etree.ElementTree.Element('testcase')

    sub_elem.set('classname', work_item.job_id)
    sub_elem.set('line', str(work_item.start_pos[0]))
    sub_elem.set('file', str(work_item.module_path))

    return sub_elem


def _update_element_with_result(sub_elem, result):
    data = result.output
    outcome = result.worker_outcome

    if outcome == WorkerOutcome.EXCEPTION:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, 'error')
        error_elem.set('message', "Worker has encountered exception")
        error_elem.text = str(data) + "\n".join(result.diff)
    elif _evaluation_success(result):
        failure_elem = xml.etree.ElementTree.SubElement(sub_elem, 'failure')
        failure_elem.set('message', "Mutant has survived your unit tests")
        failure_elem.text = str(data) + result.diff

    return sub_elem


def _evaluation_success(result):
    return result.worker_outcome == WorkerOutcome.NORMAL and \
        result.test_outcome in {TestOutcome.SURVIVED,
                                TestOutcome.INCOMPETENT}



"Tool for printing the survival rate in a session."

import math
import sys
import docopt

from cosmic_ray.work_db import use_db, WorkDB


def format_survival_rate():
    """cr-rate

Usage: cr-rate [--estimate] [--confidence <confidence_rate>] [--fail-over <max_value>] <session-file>

Calculate the survival rate of a session.

options:
    --estimate                     Print the lower bound, estimate and upper bound of
                                   survival rate
    --confidence <confidence_rate> Specify the confidence levels for estimates, 95 (%) by
                                   default. 80, 90, 95, 98, 99, 99.5, 99.8 and 99.9 are
                                   supported.
    --fail-over <max_value>        Exit with a non-zero code if the survival rate is
                                   larger than <max_value> or the calculated confidence
                                   interval is above the <max_value> (if --estimate is used).
                                   Specified as percentage.
"""
    arguments = docopt.docopt(
        format_survival_rate.__doc__, version='cr-rate 1.0')
    show_estimate = arguments['--estimate']
    confidence = arguments['--confidence']
    fail_over = arguments['--fail-over']

    # use integers as keys as equality is not well defined for floats
    # the values are z-values for Standard Normal Probabilities
    supp_z_scores = {800: 1.282,
                     900: 1.645,
                     950: 1.960,
                     980: 2.326,
                     990: 2.576,
                     995: 2.807,
                     998: 3.080,
                     999: 3.291}

    if not confidence:
        confidence = 95

    try:
        z_score = supp_z_scores[int(float(confidence)*10)]
    except KeyError:
        raise ValueError("Unsupported confidence interval: {0}"
                         .format(confidence))

    with use_db(arguments['<session-file>'], WorkDB.Mode.open) as db:
        rate = survival_rate(db)
        num_items = db.num_work_items
        num_complete = db.num_results

    if show_estimate:
        conf_int = math.sqrt(rate * (100-rate) / num_complete) \
                   * z_score * (1 - math.sqrt(num_complete / num_items))
        min_rate = rate - conf_int
        print('{:.2f} {:.2f} {:.2f}'
              .format(min_rate, rate, rate + conf_int))

    else:
        print('{:.2f}'.format(rate))
        min_rate = rate

    if fail_over and min_rate > float(fail_over):
        sys.exit(1)


def survival_rate(work_db):
    """Calcuate the survival rate for the results in a WorkDB.
    """
    kills = sum(r.is_killed for _, r in work_db.results)
    num_results = work_db.num_results

    if not num_results:
        return 0

    return (1 - kills / num_results) * 100



"A tool for generating HTML reports."

import datetime

from itertools import chain

import docopt
from yattag import Doc

from cosmic_ray.work_db import WorkDB, use_db
from cosmic_ray.work_item import TestOutcome
from cosmic_ray.tools.survival_rate import survival_rate


def report_html():
    """cr-html

Usage: cr-html [--only-completed] [--skip-success] <session-file>

Print an HTML formatted report of test results.
"""
    arguments = docopt.docopt(report_html.__doc__, version='cr-rate 1.0')
    with use_db(arguments['<session-file>'], WorkDB.Mode.open) as db:
        doc = _generate_html_report(db,
                                    arguments['--only-completed'],
                                    arguments['--skip-success'])

    print(doc.getvalue())


def _generate_html_report(db, only_completed, skip_success):
    # pylint: disable=too-many-statements
    doc, tag, text = Doc().tagtext()
    doc.asis('<!DOCTYPE html>')
    with tag('html', lang='en'):
        with tag('head'):
            doc.stag('meta', charset='utf-8')
            doc.stag(
                'meta',
                name='viewport',
                content='width=device-width, initial-scale=1, shrink-to-fit=no')
            doc.stag(
                'link',
                rel='stylesheet',
                href='https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css',
                integrity='sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T',
                crossorigin='anonymous')
            with tag('title'):
                text('Cosmic Ray Report')
        with tag('body'):
            with tag('div', klass='container'):
                with tag('h1'):
                    with tag('p', klass='text-dark'):
                        text('Cosmic Ray Report')

            all_items = db.completed_work_items
            if not only_completed:
                incomplete = ((item, None) for item in db.pending_work_items)
                all_items = chain(all_items, incomplete)

            num_items = db.num_work_items
            num_complete = db.num_results

            with tag('div', klass='container'):

                # Summary info

                with tag('div', klass='mb-1', id='summary_info___accordion'):
                    with tag('div', klass='card'):
                        with tag('a', ('data-toggle', 'collapse'), ('data-target', '#summary_info___collapse_1'),
                                 ('aria-expanded', 'true'), ('aria-controls', 'summary_info___collapse_1'), href='#'):
                            with tag('div', klass='card-header', id='summary_info___heading_1'):
                                with tag('button', klass='btn btn-outline-dark'):
                                    with tag('h4', klass='m-0'):
                                        text('Summary info')

                        with tag('div', ('aria-labelledby', 'summary_info___heading_1'),
                                 ('data-parent', '#summary_info___accordion'),
                                 klass='collapse show', id='summary_info___collapse_1'):
                            with tag('div', klass='card-body'):
                                with tag('p'):
                                    text('Date time: {}'.format(datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')))

                                with tag('p'):
                                    text('Total jobs: {}'.format(num_items))

                                if num_complete > 0:
                                    with tag('p'):
                                        text('Complete: {} ({:.2f}%)'.format(
                                            num_complete, num_complete / num_items * 100))
                                    with tag('p'):
                                        text('Survival rate: {:.2f}%'.format(survival_rate(db)))
                                else:
                                    with tag('p'):
                                        text('No jobs completed')

                # Job list

                with tag('div', klass='mb-1', id='job_list___accordion'):
                    with tag('div', klass='card'):
                        with tag('a', ('data-toggle', 'collapse'), ('data-target', '#job_list___collapse_1'),
                                 ('aria-expanded', 'false'), ('aria-controls', 'job_list___collapse_1'), href='#'):
                            with tag('div', klass='card-header', id='job_list___heading_1'):
                                with tag('button', klass='btn btn-outline-dark'):
                                    with tag('h4', klass='m-0'):
                                        text('Job list')

                        with tag('div', ('aria-labelledby', 'job_list___heading_1'),
                                 ('data-parent', '#job_list___accordion'),
                                 klass='collapse', id='job_list___collapse_1'):
                            with tag('div', klass='card-body'):
                                with tag('div', klass='text-right mb-1'):
                                    with tag('div', klass='mx-1', id='job_item_expand_all'):
                                        with tag('a',
                                                 href='#',
                                                 onclick=
                                                 '$(\'div.job_list___sub_multi_collapse\').collapse(\'show\');'):
                                            with tag('button', klass='btn btn-outline-dark'):
                                                with tag('span'):
                                                    text('Expand All')
                                    with tag('div', klass='mx-1', id='job_item_collapse_all'):
                                        with tag('a',
                                                 href='#',
                                                 onclick=
                                                 '$(\'div.job_list___sub_multi_collapse\').collapse(\'hide\');'):
                                            with tag('button', klass='btn btn-outline-dark'):
                                                with tag('span'):
                                                    text('Collapse All')

                                # Job item

                                for index, (work_item, result) in enumerate(all_items, start=1):
                                    if result is not None:
                                        if result.is_killed:
                                            if result.test_outcome == TestOutcome.INCOMPETENT:
                                                level = 'info'
                                            else:
                                                level = 'success'
                                                if skip_success:
                                                    continue
                                        else:
                                            level = 'danger'

                                    with tag('div', klass='mb-1', id='job_list___sub_accordion_{}'.format(index)):
                                        with tag('div', klass='card'):
                                            with tag('a', ('data-toggle', 'collapse'),
                                                     ('data-target', '#job_list___sub_collapse_{}_1'.format(index)),
                                                     ('aria-expanded', 'false'),
                                                     ('aria-controls', 'job_list___sub_collapse_{}_1'.format(index)),
                                                     href='#', klass='job_list___sub_multi_heading'):
                                                with tag('div', ('role', 'alert'),
                                                         klass='card-header alert-{}'.format(level),
                                                         id='job_list___sub_heading_{}_1'.format(index)):
                                                    with tag('button', klass='btn btn-outline-{}'.format(level)):
                                                        with tag('span', klass='job_id'):
                                                            text('{} : Job ID {}'.format(index, work_item.job_id))

                                            with tag('div',
                                                     ('aria-labelledby', 'job_list___sub_heading_{}_1'.format(index)),
                                                     ('data-parent', '#job_list___sub_accordion_{}'.format(index)),
                                                     klass='collapse job_list___sub_multi_collapse',
                                                     id='job_list___sub_collapse_{}_1'.format(index)):
                                                with tag('div', klass='card-body'):
                                                    with tag('div', klass='work-item'):
                                                        with tag('div',
                                                                 klass='alert alert-{} test-outcome'.format(level),
                                                                 role='alert'):
                                                            if not result.is_killed:
                                                                with tag('p'):
                                                                    text('SURVIVED')
                                                            with tag('p'):
                                                                text('worker outcome: {}'.
                                                                     format(result.worker_outcome))
                                                            with tag('p'):
                                                                text('test outcome: {}'.
                                                                     format(result.test_outcome))

                                                    with tag('pre', klass='location'):
                                                        with tag('a',
                                                                 href=pycharm_url(
                                                                     str(work_item.module_path),
                                                                     work_item.start_pos[0]), klass='text-secondary'):
                                                            with tag('button', klass='btn btn-outline-dark'):
                                                                text('{}, start pos: {}, end pos: {}'.
                                                                     format(work_item.module_path, work_item.start_pos,
                                                                            work_item.end_pos))

                                                    with tag('pre'):
                                                        text('operator: {}, occurrence: {}'.
                                                             format(work_item.operator_name, work_item.occurrence))

                                                    if result is not None:
                                                        if result.diff:
                                                            with tag('div', klass='alert alert-secondary'):
                                                                with tag('pre', klass='diff'):
                                                                    text(result.diff)

                                                        if result.output:
                                                            with tag('div', klass='alert alert-secondary'):
                                                                with tag('pre', klass='diff'):
                                                                    text(result.output)

            with tag('script'):
                doc.attr(src='https://code.jquery.com/jquery-3.3.1.slim.min.js')
                doc.attr(('integrity', 'sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo'))
                doc.attr(('crossorigin', 'anonymous'))
            with tag('script'):
                doc.attr(src='https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js')
                doc.attr(('integrity', 'sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1'))
                doc.attr(('crossorigin', 'anonymous'))
            with tag('script'):
                doc.attr(src='https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js')
                doc.attr(('integrity', 'sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM'))
                doc.attr(('crossorigin', 'anonymous'))
            with tag('script', type='text/javascript'):
                doc.asis(
                    '$(\'div.job_list___sub_multi_collapse\').on(\'shown.bs.collapse\','
                    '  function () {'
                    '    correct_behavior_functional_buttons();'
                    '  });'

                    '$(\'div.job_list___sub_multi_collapse\').on(\'hidden.bs.collapse\','
                    '  function () {'
                    '    correct_behavior_functional_buttons();'
                    '  });'

                    'function correct_behavior_functional_buttons() {'
                    '    var expand = false;'
                    '    var collapse = false;'

                    '    $(\'a.job_list___sub_multi_heading\').each(function(index) {'
                    '      if ($(this).attr(\'aria-expanded\') == \'false\') {'
                    '        expand = true;'
                    '        return false;'
                    '      };'
                    '    });'

                    '    $(\'a.job_list___sub_multi_heading\').each(function(index) {'
                    '      if ($(this).attr(\'aria-expanded\') == \'true\') {'
                    '        collapse = true;'
                    '        return false;'
                    '      };'
                    '    });'

                    '    if (expand) {'
                    '      $(\'div#job_item_expand_all\').css(\'display\', \'inline-block\');'
                    '    } else {'
                    '      $(\'div#job_item_expand_all\').css(\'display\', \'none\');'
                    '    };'

                    '    if (collapse) {'
                    '      $(\'div#job_item_collapse_all\').css(\'display\', \'inline-block\');'
                    '    } else {'
                    '      $(\'div#job_item_collapse_all\').css(\'display\', \'none\');'
                    '    };'

                    '  };'

                    'correct_behavior_functional_buttons();'
                )

    return doc


def pycharm_url(filename, line_number):
    "Get a URL for opening a file in Pycharm."
    return 'pycharm://open?file={}&line={}'.format(filename, line_number)



"Implementation of operator base class."

from abc import ABC, abstractmethod


class Operator(ABC):
    """The mutation operator base class.

    Args:
        python_version: The version of Python to use when interpreting the code in `module_path`.
            A string of the form "MAJOR.MINOR", e.g. "3.6" for Python 3.6.x.
    """

    def __init__(self, python_version):
        self._python_version = python_version

    @property
    def python_version(self):
        "Python major.minor version as a string."
        return self._python_version

    @abstractmethod
    def mutation_positions(self, node):
        """All positions where this operator can mutate `node`.

        An operator might be able to mutate a node in multiple ways, and this
        function should produce a position description for each of these
        mutations. Critically, if an operator can make multiple mutations to the
        same position, this should produce a position for each of these
        mutations (i.e. multiple identical positions).

        Returns: An iterable of `((start-line, start-col), (stop-line,
            stop-col))` tuples describing the locations where this operator will
            mutate `node`
        """

    @abstractmethod
    def mutate(self, node, index):
        """Mutate a node in an operator-specific manner.

        Return the new, mutated node. Return `None` if the node has
        been deleted. Return `node` if there is no mutation at all for
        some reason.
        """

    @classmethod
    @abstractmethod
    def examples(cls):
        """Examples of the mutations that this operator can make.

        This is primarily for testing purposes, but it could also be used for
        docmentation.

        Each example is a tuple of the form `(from-code, to-code, index)`. The
        `index` is optional and will be assumed to be 0 if it's not included.
        The `from-code` is a string containing some Python code prior to
        mutation. The `to-code` is a string desribing the code after mutation.
        `index` indicates the occurrence of the application of the operator to
        the code (i.e. for when an operator can perform multiple mutation to a
        piece of code).

        Returns: An iterable of example tuples.
        """



"Tool for printing reports on mutation testing sessions."

import docopt

from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.tools.survival_rate import survival_rate


def report():
    """cr-report

Usage: cr-report [--show-output] [--show-diff] [--show-pending] <session-file>

Print a nicely formatted report of test results and some basic statistics.

options:
    --show-output   Display output of test executions
    --show-diff     Display diff of mutants
    --show-pending  Display results for incomplete tasks
"""

    arguments = docopt.docopt(report.__doc__, version='cr-format 0.1')
    show_pending = arguments['--show-pending']
    show_output = arguments['--show-output']
    show_diff = arguments['--show-diff']

    with use_db(arguments['<session-file>'], WorkDB.Mode.open) as db:
        for work_item, result in db.completed_work_items:
            print('{} {} {} {}'.format(work_item.job_id, work_item.module_path,
                                       work_item.operator_name,
                                       work_item.occurrence))

            print('worker outcome: {}, test outcome: {}'.format(
                result.worker_outcome, result.test_outcome))

            if show_output:
                print('=== OUTPUT ===')
                print(result.output)
                print('==============')

            if show_diff:
                print('=== DIFF ===')
                print(result.diff)
                print('============')

        if show_pending:
            for work_item in db.pending_work_items:
                print('{} {} {} {}'.format(
                    work_item.job_id, work_item.module_path,
                    work_item.operator_name, work_item.occurrence))

        num_items = db.num_work_items
        num_complete = db.num_results

        print('total jobs: {}'.format(num_items))

        if num_complete > 0:
            print('complete: {} ({:.2f}%)'.format(
                num_complete, num_complete / num_items * 100))
            print('survival rate: {:.2f}%'.format(survival_rate(db)))
        else:
            print('no jobs completed')



"""Support for making clones of projects for test isolation.
"""

import contextlib
import logging
import os
from pathlib import Path
import shutil
import subprocess
import tempfile
import virtualenv

import git

from cosmic_ray.exceptions import CosmicRayTestingException as Exc

log = logging.getLogger(__name__)


@contextlib.contextmanager
def cloned_workspace(clone_config, chdir=True):
    """Create a cloned workspace and yield it.

    This creates a workspace for a with-block and cleans it up on exit. By
    default, this will also change to the workspace's `clone_dir` for the
    duration of the with-block.

    Args:
        clone_config: The execution engine configuration to use for the workspace.
        chdir: Whether to change to the workspace's `clone_dir` before entering the with-block.

    Yields: The `CloneWorkspace` instance created for the context.
    """
    workspace = ClonedWorkspace(clone_config)
    original_dir = os.getcwd()
    if chdir:
        os.chdir(workspace.clone_dir)

    try:
        yield workspace
    finally:
        os.chdir(original_dir)
        workspace.cleanup()


class ClonedWorkspace:
    """Clone a project and install it into a temporary virtual environment.

    Note that this actually *activates* the virtual environment, so don't construct one
    of these unless you want that to happen in your process.
    """

    def __init__(self, clone_config):
        self._tempdir = tempfile.TemporaryDirectory()
        log.info('New project clone in %s', self._tempdir.name)

        self._clone_dir = str(Path(self._tempdir.name) / 'repo')

        if clone_config['method'] == 'git':
            _clone_with_git(
                clone_config.get('repo-uri', '.'),
                self._clone_dir)
        elif clone_config['method'] == 'copy':
            _clone_with_copy(
                os.getcwd(),
                self._clone_dir)

        # pylint: disable=fixme
        # TODO: We should allow user to specify which version of Python to use.
        # How? The EnvBuilder could be passed a path to a python interpreter
        # which is used in the call to pip. This path would need to come from
        # the config.

        # Install into venv
        self._venv_path = Path(self._tempdir.name) / 'venv'
        log.info('Creating virtual environment in %s', self._venv_path)
        virtualenv.create_environment(str(self._venv_path))

        _activate(self._venv_path)
        _install_sitecustomize(self._venv_path)

        self._run_commands(clone_config.get('commands', ()))

    @property
    def clone_dir(self):
        "The root of the cloned project."
        return self._clone_dir

    def cleanup(self):
        "Remove the directory containin the clone and virtual environment."
        log.info('Removing temp dir %s', self._tempdir.name)
        self._tempdir.cleanup()

    def _run_commands(self, commands):
        """Run a set of commands in the workspace's virtual environment.

        Args:
            commands: An iterable of strings each representing a command to be executed.
        """
        for command in commands:
            log.info('Running installation command: %s', command)
            try:
                r = subprocess.run(command,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   shell=True,
                                   cwd=str(self._clone_dir),
                                   check=True)

                log.info('Command results: %s', r.stdout)
            except subprocess.CalledProcessError as exc:
                log.error("Error running command in virtual environment\ncommand: %s\nerror: %s",
                          command, exc.output)


def _clone_with_git(repo_uri, dest_path):
    """Create a clone by cloning a git repository.

    Args:
        repo_uri: The URI of the git repository to clone.
        dest_path: The location to clone to.
    """
    log.info('Cloning git repo %s to %s', repo_uri, dest_path)
    git.Repo.clone_from(repo_uri, dest_path, depth=1)


def _clone_with_copy(src_path, dest_path):
    """Clone a directory try by copying it.

   Args:
        src_path: The directory to be copied.
        dest_path: The location to copy the directory to.
    """
    log.info('Cloning directory tree %s to %s', src_path, dest_path)
    shutil.copytree(src_path, dest_path)


def _activate(venv_path):
    """Activate a virtual environment in the current process.

    This assumes a virtual environment that has a "activate_this.py" script, e.g.
    one created with `virtualenv` and *not* `venv`.

    Args:
        venv_path: Path of virtual environment to activate.
    """
    _home_dir, _lib_dir, _inc_dir, bin_dir = virtualenv.path_locations(str(venv_path))
    activate_script = str(Path(bin_dir) / 'activate_this.py')

    # This is the recommended way of activating venvs in a program:
    # https://virtualenv.pypa.io/en/stable/userguide/#using-virtualenv-without-bin-python
    exec(open(activate_script).read(), {'__file__': activate_script})  # pylint: disable=exec-used


_SITE_CUSTOMIZE = """
class {0}(Exception):
    pass

__builtins__['{0}'] = {0}
""".format(Exc.__name__)


def _install_sitecustomize(venv_path):
    _home_dir, lib_dir, _inc_dir, _bin_dir = virtualenv.path_locations(str(venv_path))
    with open(str(Path(lib_dir) / 'site-packages' / 'sitecustomize.py'), mode='wt', encoding='utf-8') as sc:
        sc.write(_SITE_CUSTOMIZE)



"""Implementation of the 'new-config' command.
"""

import os.path

import qprompt

from cosmic_ray.config import ConfigDict
from cosmic_ray.plugins import execution_engine_names


MODULE_PATH_HELP = """The path to the module that will be mutated.

If this is a package (as opposed to a single file module),
then all modules in the package and its subpackages will be
mutated.

This path can be absolute or relative to the location of the
config file.
"""

PYTHON_VERSION_HELP = """The version of Python to use for mutation.

If provided, this should be of the form MAJOR.MINOR. If
your mutation test runs will take place on python 3.6.4,
for example, you should specify 3.6.

If blank, then the python version to us will be detected
from the system on which the init command is run.

Generally this can be blank. You need to set it if the
Python version you're using for exec is different from
that of the workers.
"""

TEST_COMMAND_HELP = """The command to execute to run the tests on mutated code.
"""


def _validate_python_version(s):
    "Return True if a string is of the form <int>.<int>, False otherwise."
    if not s:
        return True
    toks = s.split('.')
    if len(toks) != 2:
        return False
    try:
        int(toks[0])
        int(toks[1])
    except ValueError:
        return False
    return True


def new_config():
    """Prompt user for config variables and generate new config.

    Returns: A new ConfigDict.
    """
    config = ConfigDict()
    config["module-path"] = qprompt.ask_str(
        "Top-level module path",
        blk=False,
        vld=os.path.exists,
        hlp=MODULE_PATH_HELP)

    python_version = qprompt.ask_str(
        'Python version (blank for auto detection)',
        vld=_validate_python_version,
        hlp=PYTHON_VERSION_HELP)
    config['python-version'] = python_version

    timeout = qprompt.ask_str(
        'Test execution timeout (seconds)',
        vld=float,
        blk=False,
        hlp="The number of seconds to let a test run before terminating it.")
    config['timeout'] = float(timeout)
    config['excluded-modules'] = []

    config["test-command"] = qprompt.ask_str(
        "Test command",
        blk=False,
        hlp=TEST_COMMAND_HELP)

    menu = qprompt.Menu()
    for at_pos, engine_name in enumerate(execution_engine_names()):
        menu.add(str(at_pos), engine_name)
    config["execution-engine"] = ConfigDict()
    config['execution-engine']['name'] = menu.show(header="Execution engine", returns="desc")

    config["cloning"] = ConfigDict()
    config['cloning']['method'] = 'copy'
    config['cloning']['commands'] = []

    return config



"""Cosmic Ray execution engine that runs mutations locally on git clones.

This engine creates a pool of subprocesses, each of which execute some portion
of work items in a session. Each of these processes creates a shallow clone of
the git repository containing the code to be mutated/tested in a temporary
directory. This way each process can work independently.

## Enabling the engine

To use the local-git execution engine, set `cosmic-ray.execution-engine.name =
"local-git"` in your Cosmic Ray configuration::

    [cosmic-ray.execution-engine]
    name = "local-git"

## Specifying the source repository

Each subprocess creates its own clone of a source git repository. By default,
the engine determines which repo to use by looking for the repo dominating the
directory in which `cosmic-ray` is executed. However, you can specify a
different repository (e.g. a repo hosted on github) by setting the
`cosmic-ray.execution-engine.local-git.repo-uri`::

    [cosmic-ray.execution-engine.local-git]
    repo-uri = "https://github.com/me/difference-engine-emulator"

## Subprocess environment

The local-git engine launches its subprocesses using the `multiprocessing`
library. As such, the subprocesses run in an environment that is largely
identical to that of the main `cosmic-ray` process.

One major difference is the directory in which the subprocesses run. They run
the root of the cloned repository, so you need to take this into account when
creating the configuration.
"""

import contextlib
import logging
import multiprocessing
import multiprocessing.util
import os

from cosmic_ray.cloning import ClonedWorkspace
from cosmic_ray.execution.execution_engine import ExecutionEngine
from cosmic_ray.worker import worker

log = logging.getLogger(__name__)

# Per-subprocess globals
_workspace = None
_config = None


@contextlib.contextmanager
def excursion(dirname):
    "Context manager for temporarily changing directories."
    orig = os.getcwd()
    try:
        os.chdir(dirname)
        yield
    finally:
        os.chdir(orig)


def _initialize_worker(config):
    # pylint: disable=global-statement
    global _workspace
    global _config
    assert _workspace is None
    assert _config is None

    _config = config

    log.info('Initialize local-git worker in PID %s', os.getpid())
    _workspace = ClonedWorkspace(config.cloning_config)

    # Register a finalizer
    multiprocessing.util.Finalize(_workspace, _workspace.cleanup, exitpriority=16)


def _execute_work_item(work_item):
    log.info('Executing worker in %s, PID=%s', _workspace.clone_dir, os.getpid())

    with excursion(_workspace.clone_dir):
        result = worker(
            work_item.module_path,
            _config.python_version,
            work_item.operator_name,
            work_item.occurrence,
            _config.test_command,
            _config.timeout)

    return work_item.job_id, result


class LocalExecutionEngine(ExecutionEngine):
    "The local-git execution engine."

    def __call__(self, pending_work, config, on_task_complete):
        with multiprocessing.Pool(
                initializer=_initialize_worker,
                initargs=(config,)) as pool:

            # pylint: disable=W0511
            # TODO: This is not optimal. The pending-work iterable could be millions
            # or billions of elements. We don't want to copy it. We copy it right
            # now so that we don't access the database in a separate thread (i.e.
            # one created by imap_unoredered below). We need to find a way around
            # this.
            pending = list(pending_work)

            results = pool.imap_unordered(
                func=_execute_work_item,
                iterable=pending)

            for job_id, result in results:
                on_task_complete(job_id, result)



"Base execution-engine implementation details."

import abc


class ExecutionEngine(metaclass=abc.ABCMeta):
    "Base class for execution engine plugins."

    @abc.abstractmethod
    def __call__(self, pending_work, config, on_task_complete):
        """Execute jobs in `pending_work_items`.

        Spend no more than `config.timeout` seconds for a single job, using `config` to
        control the work.
        """



"""An filter that removes operators based on regular expressions.
"""
from argparse import Namespace
import logging
import re
import sys

from cosmic_ray.config import load_config
from cosmic_ray.work_db import WorkDB
from cosmic_ray.work_item import WorkerOutcome, WorkResult
from cosmic_ray.tools.filters.filter_app import FilterApp

log = logging.getLogger()


class OperatorsFilter(FilterApp):
    "Implemenents the operators-filter."
    def description(self):
        return __doc__

    def _skip_filtered(self, work_db, exclude_operators):
        re_exclude_operators = re.compile('|'.join('(:?%s)' % e for e in exclude_operators))

        for item in work_db.pending_work_items:
            if re_exclude_operators.match(item.operator_name):
                log.info(
                    "operator skipping %s %s %s %s %s %s",
                    item.job_id,
                    item.operator_name,
                    item.occurrence,
                    item.module_path,
                    item.start_pos,
                    item.end_pos,
                )

                work_db.set_result(
                    item.job_id,
                    WorkResult(
                        output="Filtered operator",
                        worker_outcome=WorkerOutcome.SKIPPED,
                    ),
                )

    def filter(self, work_db: WorkDB, args: Namespace):
        """Mark as skipped all work item with filtered operator
        """

        if args.config is None:
            config = work_db.get_config()
        else:
            config = load_config(args.config)

        exclude_operators = config.sub('filters', 'operators-filter').get('exclude-operators', ())
        self._skip_filtered(work_db, exclude_operators)

    def add_args(self, parser):
        parser.add_argument('--config', help='Config file to use')


def main(argv=None):
    """Run the operators-filter with the specified command line arguments.
    """
    return OperatorsFilter().main(argv)


if __name__ == '__main__':
    sys.exit(main())



"""Implementation of the WorkDB."""

import contextlib
import os
import sqlite3
from enum import Enum

from .config import deserialize_config, serialize_config
from .work_item import TestOutcome, WorkerOutcome, WorkItem, WorkResult


class WorkDB:
    """WorkDB is the database that keeps track of mutation testing work progress.

    Essentially, there's a row in the DB for each mutation that needs to be
    executed in some run. These initially start off with no results, and
    results are added as they're completed.
    """

    class Mode(Enum):
        "Modes in which a WorkDB may be opened."

        # Open existing files, creating if necessary
        create = 1

        # Open only existing files, failing if it doesn't exist
        open = 2

    def __init__(self, path, mode):
        """Open a DB in file `path` in mode `mode`.

        Args:
          path: The path to the DB file.
          mode: The mode in which to open the DB. See the `Mode` enum for
            details.

        Raises:
          FileNotFoundError: If `mode` is `Mode.open` and `path` does not
            exist.
        """

        if (mode == WorkDB.Mode.open) and (not os.path.exists(path)):
            raise FileNotFoundError('Corresponding database {} not found'.format(path))

        self._path = path
        self._conn = sqlite3.connect(str(path))

        self._init_db()

    def close(self):
        """Close the database."""
        self._conn.close()

    @property
    def name(self):
        """A name for this database.

        Derived from the constructor arguments.
        """
        return self._path

    def set_config(self, config):
        """Set (replace) the configuration for the session.

        Args:
          config: Configuration object
        """
        with self._conn:
            self._conn.execute("DELETE FROM config")
            self._conn.execute('INSERT INTO config VALUES(?)',
                               (serialize_config(config),))

    def get_config(self):
        """Get the work parameters (if set) for the session.

        Returns: a Configuration object.

        Raises:
          ValueError: If is no config set for the session.
        """
        rows = list(self._conn.execute("SELECT * FROM config"))
        if not rows:
            raise ValueError("work-db has no config")
        (config_str,) = rows[0]

        return deserialize_config(config_str)

    @property
    def work_items(self):
        """An iterable of all of WorkItems in the db.

        This includes both WorkItems with and without results.
        """
        cur = self._conn.cursor()
        rows = cur.execute("SELECT * FROM work_items")
        for row in rows:
            yield _row_to_work_item(row)

    @property
    def num_work_items(self):
        """The number of work items."""
        count = self._conn.execute("SELECT COUNT(*) FROM work_items")
        return list(count)[0][0]

    def add_work_item(self, work_item):
        """Add a WorkItems.

        Args:
          work_item: A WorkItem.
        """
        with self._conn:
            self._conn.execute(
                '''
                INSERT INTO work_items
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                ''', _work_item_to_row(work_item))

    def add_work_items(self, work_items):
        """Add multiple WorkItems.

        Unlike calling `add_work_item` multiple times, performs all insertions
        in a single transaction.

        Args:
          work_items: an iterable of WorkItem.
        """
        with self._conn:
            self._conn.execute('BEGIN TRANSACTION')
            for w_i in work_items:
                self._conn.execute(
                    '''
                    INSERT INTO work_items
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                    ''', _work_item_to_row(w_i))
            if self._conn.isolation_level:
                self._conn.execute('END TRANSACTION')

    def clear(self):
        """Clear all work items from the session.

        This removes any associated results as well.
        """
        with self._conn:
            self._conn.execute('DELETE FROM results')
            self._conn.execute('DELETE FROM work_items')

    @property
    def results(self):
        "An iterable of all `(job-id, WorkResult)`s."
        cur = self._conn.cursor()
        rows = cur.execute("SELECT * FROM results")
        for row in rows:
            yield (row['job_id'], _row_to_work_result(row))

    @property
    def num_results(self):
        """The number of results."""
        count = self._conn.execute("SELECT COUNT(*) FROM results")
        return list(count)[0][0]

    def set_result(self, job_id, result):
        """Set the result for a job.

        This will overwrite any existing results for the job.

        Args:
          job_id: The ID of the WorkItem to set the result for.
          result: A WorkResult indicating the result of the job.

        Raises:
           KeyError: If there is no work-item with a matching job-id.
        """
        with self._conn:
            try:
                self._conn.execute(
                    '''
                    REPLACE INTO results
                    VALUES (?, ?, ?, ?, ?)
                    ''', _work_result_to_row(job_id, result))
            except sqlite3.IntegrityError as exc:
                raise KeyError('Can not add result with job-id {}'.format(
                    job_id)) from exc

    @property
    def pending_work_items(self):
        "Iterable of all pending work items. In random order."
        pending = self._conn.execute(
            "SELECT * FROM work_items "
            "WHERE job_id NOT IN (SELECT job_id FROM results) "
            "ORDER BY hex(randomblob(16))"
        )
        return (_row_to_work_item(p) for p in pending)

    @property
    def completed_work_items(self):
        "Iterable of `(work-item, result)`s for all completed items."
        completed = self._conn.execute(
            "SELECT * FROM work_items, results WHERE work_items.job_id == results.job_id"
        )
        return ((_row_to_work_item(result), _row_to_work_result(result))
                for result in completed)

    # @property
    # def num_pending_work_items(self):
    #     "The number of pending WorkItems in the session."
    #     count = self._conn.execute("SELECT COUNT(*) FROM work_items WHERE job_id NOT IN (SELECT job_id FROM results)")
    #     return count[0][0]

    def _init_db(self):
        with self._conn:
            self._conn.row_factory = sqlite3.Row

            self._conn.execute("PRAGMA foreign_keys = 1")

            # journal_mode=WAL is persistent
            self._conn.execute("PRAGMA journal_mode=WAL")

            self._conn.execute('''
            CREATE TABLE IF NOT EXISTS work_items
            (module_path text,
             operator text,
             occurrence int,
             start_line int,
             start_col int,
             end_line int,
             end_col int,
             job_id text primary key)
            ''')

            self._conn.execute('''
            CREATE TABLE IF NOT EXISTS results
            (worker_outcome text,
             output text,
             test_outcome text,
             diff text,
             job_id text primary key,
             FOREIGN KEY(job_id) REFERENCES work_items(job_id)
            )
            ''')

            self._conn.execute('''
            CREATE TABLE IF NOT EXISTS config
            (config text)
            ''')


def _row_to_work_item(row):
    return WorkItem(
        module_path=row['module_path'],
        operator_name=row['operator'],
        occurrence=row['occurrence'],
        start_pos=(row['start_line'], row['start_col']),
        end_pos=(row['end_line'], row['end_col']),
        job_id=row['job_id'])


def _work_item_to_row(work_item):
    return (
        str(work_item.module_path),
        work_item.operator_name,
        work_item.occurrence,
        work_item.start_pos[0],
        work_item.start_pos[1],
        work_item.end_pos[0],
        work_item.end_pos[1],
        work_item.job_id)


def _row_to_work_result(row):
    test_outcome = row['test_outcome']
    test_outcome = None if test_outcome is None else TestOutcome(test_outcome)

    return WorkResult(
        worker_outcome=WorkerOutcome(row['worker_outcome']),
        output=row['output'],
        test_outcome=test_outcome,
        diff=row['diff'])


def _work_result_to_row(job_id, result):
    return (
        result.worker_outcome.value,  # should never be None
        result.output,
        None if result.test_outcome is None else result.test_outcome.value,
        result.diff,
        job_id)


@contextlib.contextmanager
def use_db(path, mode=WorkDB.Mode.create):
    """
    Open a DB in file `path` in mode `mode` as a context manager.

    On exiting the context the DB will be automatically closed.

    Args:
      path: The path to the DB file.
      mode: The mode in which to open the DB. See the `Mode` enum for
        details.

    Raises:
      FileNotFoundError: If `mode` is `Mode.open` and `path` does not
        exist.
    """
    database = WorkDB(path, mode)
    try:
        yield database

    finally:
        database.close()



"""Tool for creating badge.
"""
import os
from logging import getLogger

from anybadge import Badge

import docopt
from cosmic_ray.config import load_config
from cosmic_ray.tools.survival_rate import survival_rate
from cosmic_ray.work_db import WorkDB, use_db

log = getLogger()


def generate_badge():
    """cr-badge

Usage: cr-badge [--config <config_file>]  <badge_file> <session-file>

Generate badge file.

options:
    --config <config_file> Configuration file to use instead of session configuration
"""

    arguments = docopt.docopt(generate_badge.__doc__, version='cr-format 0.1')
    config_file = arguments['--config']
    badge_filename = arguments['<badge_file>']

    with use_db(arguments['<session-file>'], WorkDB.Mode.open) as db:
        assert isinstance(db, WorkDB)
        if config_file:
            config = load_config(config_file)
        else:
            config = db.get_config()

        percent = 100 - survival_rate(db)

        config = config['badge']

        badge = Badge(
            label=config['label'],
            value=percent,
            value_format=config['format'],
            thresholds=config['thresholds'],
        )

        log.info("Generating badge: " + config['format'], percent)  # pylint: disable=logging-not-lazy

        try:
            os.unlink(badge_filename)
        except OSError:
            pass

        badge.write_badge(badge_filename)



"""A filter that uses "# pragma: no mutate" to determine when specific mutations
should be skipped.
"""
import logging
import re
import sys
from functools import lru_cache

from cosmic_ray.tools.filters.filter_app import FilterApp
from cosmic_ray.work_item import WorkerOutcome, WorkResult

log = logging.getLogger()


class PragmaNoMutateFilter(FilterApp):
    """Implements the pragma-no-mutate filter.
    """
    def description(self):
        return __doc__

    def filter(self, work_db, _args):
        """Mark lines with "# pragma: no mutate" as SKIPPED

        For all work_item in db, if the LAST line of the working zone is marked
        with "# pragma: no mutate", This work_item will be skipped.
        """

        @lru_cache()
        def file_contents(file_path):
            "A simple cache of file contents."
            with file_path.open(mode="rt") as handle:
                return handle.readlines()

        re_is_mutate = re.compile(r'.*#.*pragma:.*no mutate.*')

        for item in work_db.work_items:
            lines = file_contents(item.module_path)
            try:
                # item.{start,end}_pos[0] seems to be 1-based.
                line_number = item.end_pos[0] - 1
                if item.end_pos[1] == 0:
                    # The working zone ends at begin of line,
                    # consider the previous line.
                    line_number -= 1
                line = lines[line_number]
                if re_is_mutate.match(line):
                    work_db.set_result(item.job_id,
                                       WorkResult(output=None,
                                                  test_outcome=None,
                                                  diff=None,
                                                  worker_outcome=WorkerOutcome.SKIPPED))
            except Exception as ex:
                raise Exception("module_path: %s, start_pos: %s, end_pos: %s, len(lines): %s" %
                                (item.module_path, item.start_pos, item.end_pos, len(lines))) from ex


def main(argv=None):
    """Run pragma-no-mutate filter with specified command line arguments.
    """
    return PragmaNoMutateFilter().main(argv)


if __name__ == '__main__':
    sys.exit(main())



"""A filter that uses spor metadata to determine when specific mutations
should be skipped.

Spor anchors that include `{mutate: False}` will cause mutations at that location to be skipped.
"""
from functools import lru_cache
import logging
import sys

from spor.repository import open_repository

from cosmic_ray.tools.filters.filter_app import FilterApp
from cosmic_ray.work_item import WorkerOutcome, WorkResult

log = logging.getLogger()


class SporFilter(FilterApp):
    """Implementation of spor filter.
    """
    def description(self):
        return __doc__

    def filter(self, work_db, _args):
        """Look for WorkItems in `work_db` that should not be mutated due to spor metadata.

        For each WorkItem, find anchors for the item's file/line/columns. If an
        anchor exists with metadata containing `{mutate: False}` then the WorkItem
        is marked as SKIPPED.

        Args:
            work_db: A WorkDB to filter.
        """

        @lru_cache()
        def file_contents(file_path):
            "A simple cache of file contents."
            with file_path.open(mode="rt") as handle:
                return handle.readlines()

        for item in work_db.pending_work_items:
            try:
                repo = open_repository(item.module_path)
            except ValueError:
                log.info("No spor repository for %s", item.module_path)
                continue

            for _, anchor in repo.items():
                if anchor.file_path != item.module_path.absolute():
                    continue

                metadata = anchor.metadata

                lines = file_contents(item.module_path)
                if _item_in_context(
                        lines, item,
                        anchor.context) and not metadata.get("mutate", True):
                    log.info(
                        "spor skipping %s %s %s %s %s %s",
                        item.job_id,
                        item.operator_name,
                        item.occurrence,
                        item.module_path,
                        item.start_pos,
                        item.end_pos,
                    )

                    work_db.set_result(
                        item.job_id,
                        WorkResult(
                            output="Filtered by spor",
                            test_outcome=None,
                            diff=None,
                            worker_outcome=WorkerOutcome.SKIPPED,
                        ),
                    )


def main(argv=None):
    """Run spor filter with specified command line args.

    Args:
        argv: Command-line arguments to use.
    """
    return SporFilter().main(argv)


if __name__ == '__main__':
    sys.exit(main())


def _line_and_col_to_offset(lines, line, col):
    """Figure out the offset into a file for a particular line and col.

    This can return offsets that don't actually exist in the file. If you
    specify a line that exists and a col that is past the end of that line, this
    will return a "fake" offset. This is to account for the fact that a
    WorkItem's end_pos is one-past the end of a mutation, and hence potentially
    one-past the end of a file.

    Args:
        lines: A sequence of the lines in a file.
        line: A one-based index indicating the line in the file.
        col: A zero-based index indicating the column on `line`.

    Raises: ValueError: If the specified line found in the file.
    """

    offset = 0
    for index, contents in enumerate(lines, 1):
        if index == line:
            return offset + col

        offset += len(contents)

    raise ValueError("Offset {}:{} not found".format(line, col))


def _item_in_context(lines, item, context):
    """Determines if a WorkItem falls within an anchor.

    This only returns True if a WorkItems start-/stop-pos range is *completely*
    within an anchor, not just if it overalaps.
    """
    start_offset = _line_and_col_to_offset(lines, item.start_pos[0],
                                           item.start_pos[1])
    stop_offset = _line_and_col_to_offset(lines, item.end_pos[0],
                                          item.end_pos[1])
    width = stop_offset - start_offset

    return start_offset >= context.offset and width <= len(context.topic)



"Implementation of the 'init' command."
import logging
import uuid

from cosmic_ray.ast import get_ast, ast_nodes
import cosmic_ray.modules
from cosmic_ray.work_item import WorkItem
from cosmic_ray.plugins import get_operator
from cosmic_ray.work_db import WorkDB

log = logging.getLogger()


def all_work_items(module_paths, operator_names, python_version):
    "Iterable of all WorkItems for the given inputs."
    for module_path in module_paths:
        module_ast = get_ast(
            module_path, python_version=python_version)


        for op_name in operator_names:
            operator = get_operator(op_name)(python_version)
            occurrence = 0
            for node in ast_nodes(module_ast):
                for start_pos, end_pos in operator.mutation_positions(node):
                    yield WorkItem(
                        job_id=uuid.uuid4().hex,
                        module_path=str(module_path),
                        operator_name=op_name,
                        occurrence=occurrence,
                        start_pos=start_pos,
                        end_pos=end_pos)

                    occurrence += 1


def init(module_paths, work_db: WorkDB, config):
    """Clear and initialize a work-db with work items.

    Any existing data in the work-db will be cleared and replaced with entirely
    new work orders. In particular, this means that any results in the db are
    removed.

    Args:
      module_paths: iterable of pathlib.Paths of modules to mutate.
      work_db: A `WorkDB` instance into which the work orders will be saved.
      config: The configuration for the new session.
    """

    operator_names = list(cosmic_ray.plugins.operator_names())

    work_db.set_config(config=config)

    work_db.clear()
    work_db.add_work_items(
        all_work_items(
            module_paths, 
            operator_names, 
            config.python_version))



"""Support for making mutations to source code.
"""
from contextlib import contextmanager

from cosmic_ray.ast import get_ast, Visitor


@contextmanager
def use_mutation(module_path, operator, occurrence):
    """A context manager that applies a mutation for the duration of a with-block.

    This applies a mutation to a file on disk, and after the with-block it put the unmutated code
    back in place.

    Args:
        module_path: The path to the module to mutate.
        operator: The `Operator` instance to use.
        occurrence: The occurrence of the operator to apply.

    Yields: A `(unmutated-code, mutated-code)` tuple to the with-block. If there was
        no mutation performed, the `mutated-code` is `None`.
    """
    original_code, mutated_code = apply_mutation(module_path, operator,
                                                 occurrence)
    try:
        yield original_code, mutated_code
    finally:
        with module_path.open(mode='wt', encoding='utf-8') as handle:
            handle.write(original_code)
            handle.flush()


def apply_mutation(module_path, operator, occurrence):
    """Apply a specific mutation to a file on disk.

    Args:
        module_path: The path to the module to mutate.
        operator: The `operator` instance to use.
        occurrence: The occurrence of the operator to apply.

    Returns: A `(unmutated-code, mutated-code)` tuple to the with-block. If there was
        no mutation performed, the `mutated-code` is `None`.
    """
    module_ast = get_ast(module_path, python_version=operator.python_version)
    original_code = module_ast.get_code()
    visitor = MutationVisitor(occurrence, operator)
    mutated_ast = visitor.walk(module_ast)

    mutated_code = None
    if visitor.mutation_applied:
        mutated_code = mutated_ast.get_code()
        with module_path.open(mode='wt', encoding='utf-8') as handle:
            handle.write(mutated_code)
            handle.flush()

    return original_code, mutated_code


class MutationVisitor(Visitor):
    """Visitor that mutates a module with the specific occurrence of an operator.

    This will perform at most one mutation in a walk of an AST. If this performs
    a mutation as part of the walk, it will store the mutated node in the
    `mutant` attribute. If the walk does not result in any mutation, `mutant`
    will be `None`.

    Note that `mutant` is just the specifically mutated node. It will generally
    be a part of the larger AST which is returned from `walk()`.
    """

    def __init__(self, occurrence, operator):
        self.operator = operator
        self._occurrence = occurrence
        self._count = 0
        self._mutation_applied = False

    @property
    def mutation_applied(self):
        "Whether this visitor has applied a mutation."
        return self._mutation_applied

    def visit(self, node):
        for index, _ in enumerate(self.operator.mutation_positions(node)):
            if self._count == self._occurrence:
                self._mutation_applied = True
                node = self.operator.mutate(node, index)
            self._count += 1

        return node



"""Classes for describing work and results.
"""
import enum
import json
import pathlib


class StrEnum(str, enum.Enum):
    "An Enum subclass with str values."


class WorkerOutcome(StrEnum):
    """Possible outcomes for a worker.
    """
    NORMAL = 'normal'  # The worker exited normally, producing valid output
    EXCEPTION = 'exception'  # The worker exited with an exception
    ABNORMAL = 'abnormal'  # The worker did not exit normally or with an exception (e.g. a segfault)
    NO_TEST = 'no-test'  # The worker had no test to run
    SKIPPED = 'skipped'  # The job was skipped (worker was not executed)


class TestOutcome(StrEnum):
    """A enum of the possible outcomes for any mutant test run.
    """
    SURVIVED = 'survived'
    KILLED = 'killed'
    INCOMPETENT = 'incompetent'


class WorkResult:
    """The result of a single mutation and test run.
    """

    def __init__(self,
                 worker_outcome,
                 output=None,
                 test_outcome=None,
                 diff=None):
        if worker_outcome is None:
            raise ValueError('Worker outcome must always have a value.')

        self._output = output
        self._test_outcome = test_outcome
        self._worker_outcome = worker_outcome
        self._diff = diff

    @property
    def worker_outcome(self):
        "A `WorkerOutcome` indicating how the worker finished."
        return self._worker_outcome

    @property
    def test_outcome(self):
        "A `TestOutcome` indicating how the test runner finished. Possibly `None`."
        return self._test_outcome

    @property
    def output(self):
        "Any output returned by the test command. Possibly `None`."
        return self._output

    @property
    def diff(self):
        "A sequence of strings containing the diff generated by the mutation. Possibly `None`."
        return self._diff

    def as_dict(self):
        "Get the WorkResult as a dict."
        return {
            'output': self.output,
            'test_outcome': self.test_outcome,
            'worker_outcome': self.worker_outcome,
            'diff': self.diff,
        }

    @property
    def is_killed(self):
        "Whether the mutation should be considered 'killed'"
        return self.test_outcome != TestOutcome.SURVIVED

    def __eq__(self, rhs):
        return self.as_dict() == rhs.as_dict()

    def __neq__(self, rhs):
        return not self == rhs

    def __repr__(self):
        return "<WorkResult {test_outcome}/{worker_outcome}: '{output}'>".format(
            test_outcome=self._test_outcome,
            worker_outcome=self.worker_outcome,
            output=self.output)


class WorkItem:
    """Description of the work for a single mutation and test run.
    """

    # pylint: disable=R0913
    def __init__(self,
                 module_path=None,
                 operator_name=None,
                 occurrence=None,
                 start_pos=None,
                 end_pos=None,
                 job_id=None):
        if start_pos[0] > end_pos[0]:
            raise ValueError('Start line must not be after end line')

        if start_pos[0] == end_pos[0]:
            if start_pos[1] >= end_pos[1]:
                raise ValueError(
                    'End position must come after start position.')

        self._module_path = pathlib.Path(module_path)
        self._operator_name = operator_name
        self.occurrence = occurrence
        self._start_pos = start_pos
        self._end_pos = end_pos
        self._job_id = job_id

    @property
    def module_path(self):
        "pathlib.Path to module being mutated."
        return self._module_path

    @property
    def operator_name(self):
        "The name of the operator (i.e. as defined by the provider)"
        return self._operator_name

    @property
    def start_pos(self):
        "Start of the mutation location as a `(line, column)` tuple."
        return self._start_pos

    @property
    def end_pos(self):
        """End of the mutation location as a `(line, column)` tuple.

        Note that this represents the offset *one past* the end of the mutated
        segment. If the mutated segment is at the end of a file, this offset
        will be past the end of the file.
        """
        return self._end_pos

    @property
    def job_id(self):
        "The unique ID of the job"
        return self._job_id

    def as_dict(self):
        """Get fields as a dict.
        """
        return {
            'module_path': str(self.module_path),
            'operator_name': self.operator_name,
            'occurrence': self.occurrence,
            'start_pos': self.start_pos,
            'end_pos': self.end_pos,
            'job_id': self.job_id,
        }

    def __eq__(self, rhs):
        return self.as_dict() == rhs.as_dict()

    def __neq__(self, rhs):
        return not self == rhs

    def __repr__(self):
        return "<WorkItem {job_id}: ({start_pos}/{end_pos}) {occurrence} - {operator} ({module})>".format(
            job_id=self.job_id,
            start_pos=self.start_pos,
            end_pos=self.end_pos,
            occurrence=self.occurrence,
            operator=self.operator_name,
            module=self.module_path)


class WorkItemJsonEncoder(json.JSONEncoder):
    "Custom JSON encoder for workitems and workresults."

    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, WorkItem):
            return {"_type": "WorkItem", "values": o.as_dict()}

        if isinstance(o, WorkResult):
            return {"_type": "WorkResult", "values": o.as_dict()}

        return super().default(o)


class WorkItemJsonDecoder(json.JSONDecoder):
    "Custom JSON decoder for WorkItems and WorkResults."

    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=self._decode_work_items)

    @staticmethod
    def _decode_work_items(obj):
        if (obj.get('_type') == 'WorkItem') and ('values' in obj):
            values = obj['values']
            return WorkItem(**values)

        if (obj.get('_type') == 'WorkResult') and ('values' in obj):
            values = obj['values']
            return WorkResult(**values)

        return obj



"""Setup for Cosmic Ray.
"""

from pathlib import Path

from setuptools import setup, find_packages


INSTALL_REQUIRES = [
    'astunparse',
    'click',
    'decorator',
    'docopt',
    'exit_codes',
    'gitpython',
    'parso',
    'pathlib',
    'qprompt',
    'spor>=1.1.0',
    'stevedore',
    'toml',
    'virtualenv<=16.7.10',
    'yattag',
    'anybadge',
]

setup(
    name='cosmic_ray',
    version='6.1.1',
    packages=find_packages('src'),
    author='Sixty North AS',
    author_email='austin@sixty-north.com',
    description='Mutation testing',
    license='MIT License',
    keywords='testing',
    package_dir={'': 'src'},
    url='http://github.com/sixty-north/cosmic-ray',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Testing',
    ],
    platforms='any',
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'test': ['hypothesis', 'pytest', 'pytest-mock', 'tox'],
        'dev': ['pylint', 'autopep8', 'bumpversion'],
        'docs': ['sphinx', 'sphinx_rtd_theme'],
        'celery4_engine': ['cosmic_ray_celery4_engine'],
    },
    entry_points={
        'console_scripts': [
            'cosmic-ray = cosmic_ray.cli:main',
            'cr-html = cosmic_ray.tools.html:report_html',
            'cr-report = cosmic_ray.tools.report:report',
            'cr-badge = cosmic_ray.tools.badge:generate_badge',
            'cr-rate = cosmic_ray.tools.survival_rate:format_survival_rate',
            'cr-xml = cosmic_ray.tools.xml:report_xml',
            'cr-filter-spor = cosmic_ray.tools.filters.spor:main',
            'cr-filter-operators = cosmic_ray.tools.filters.operators_filter:main',
            'cr-filter-pragma = cosmic_ray.tools.filters.pragma_no_mutate:main',
            'cr-filter-git = cosmic_ray.tools.filters.git:main',
        ],
        'cosmic_ray.test_runners': [
            'unittest = cosmic_ray.testing.unittest_runner:UnittestRunner',
        ],
        'cosmic_ray.operator_providers': [
            'core = cosmic_ray.operators.provider:OperatorProvider',
        ],
        'cosmic_ray.execution_engines': [
            'local = cosmic_ray.execution.local:LocalExecutionEngine',

        ],
    },
    long_description=Path('README.rst').read_text(encoding='utf-8'),
)



"Implementation of the boolean replacement operators."

import parso.python.tree

from .keyword_replacer import KeywordReplacementOperator
from .operator import Operator


class ReplaceTrueWithFalse(KeywordReplacementOperator):
    """An that replaces True with False."""
    from_keyword = 'True'
    to_keyword = 'False'


class ReplaceFalseWithTrue(KeywordReplacementOperator):
    """An that replaces False with True."""
    from_keyword = 'False'
    to_keyword = 'True'


class ReplaceAndWithOr(KeywordReplacementOperator):
    """An operator that swaps 'and' with 'or'."""
    from_keyword = 'and'
    to_keyword = 'or'

    @classmethod
    def examples(cls):
        return (
            ('x and y', 'x or y'),
        )


class ReplaceOrWithAnd(KeywordReplacementOperator):
    """An operator that swaps 'or' with 'and'."""
    from_keyword = 'or'
    to_keyword = 'and'

    @classmethod
    def examples(cls):
        return (
            ('x or y', 'x and y'),
        )


class AddNot(Operator):
    """
        An operator that adds the 'not' keyword to boolean expressions.

        NOTE: 'not' as unary operator is mutated in
         `unary_operator_replacement.py`, including deletion of the same
         operator.
    """
    NODE_TYPES = (parso.python.tree.IfStmt, parso.python.tree.WhileStmt,
                  parso.python.tree.AssertStmt)

    def mutation_positions(self, node):
        if isinstance(node, self.NODE_TYPES):
            expr = node.children[1]
            yield (expr.start_pos, expr.end_pos)
        elif isinstance(node,
                        parso.python.tree.PythonNode) and node.type == 'test':
            # ternary conditional
            expr = node.children[2]
            yield (expr.start_pos, expr.end_pos)

    def mutate(self, node, index):
        assert index == 0

        if isinstance(node, self.NODE_TYPES):
            expr_node = node.children[1]
            mutated_code = ' not{}'.format(expr_node.get_code())
            mutated_node = parso.parse(mutated_code)
            node.children[1] = mutated_node

        else:
            assert node.type == 'test'
            expr_node = node.children[2]
            mutated_code = ' not{}'.format(expr_node.get_code())
            mutated_node = parso.parse(mutated_code)
            node.children[2] = mutated_node

        return node

    @classmethod
    def examples(cls):
        return (
            ('if True or False: pass', 'if not True or False: pass'),
            ('A if B else C', 'A if not B else C'),
            ('assert isinstance(node, ast.Break)', 'assert not isinstance(node, ast.Break)'),
            ('while True: pass', 'while not True: pass'),
        )
